//#####################################################################
//  Copyright (c) 2011-2015 Anonymous Authors of Paper Submission paper0191.
//  This file is covered by the FreeBSD license. Please refer to the 
//  license.txt file for more information.
//#####################################################################


#ifndef __KERNEL_NUMBER_MIC_H__
#define __KERNEL_NUMBER_MIC_H__


//==============================================================//
//                                                              //
//                      CONSTRUCTORS                            //
//                                                              //
//==============================================================//


template<> inline
Number<__m512>::Number()
{value=_mm512_castsi512_ps(_mm512_xor_epi32(_mm512_castps_si512(value),_mm512_castps_si512(value)));}



//==============================================================//
//==============================================================//

//==============================================================//
//                                                              //
//                      BASIC OPERATIONS                        //
//                                                              //
//==============================================================//


//------------------------------------//
//             ADDITION               //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::operator+(const Number& other) const
{Number<__m512> result;result.value=_mm512_add_ps(value,other.value);return result;}



//------------------------------------//
//           MULTIPLICATION           //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::operator*(const Number& other) const
{Number<__m512> result;result.value=_mm512_mul_ps(value,other.value);return result;}


//------------------------------------//
//           SUBTRACTION              //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::operator-(const Number& other) const
{Number<__m512> result;result.value=_mm512_sub_ps(value,other.value);return result;}


//------------------------------------//
//            DIVISION                //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::operator/(const Number& other) const
{Number<__m512> result;result.value=_mm512_div_ps(value,other.value);return result;}


//==============================================================//
//                                                              //
//                      COMPARISON OPERATIONS                   //
//                                                              //
//==============================================================//

//------------------------------------//
//             LESS THAN              //
//------------------------------------//


template<> inline
typename Number<__m512>::Mask Number<__m512>::operator<(const Number& other) const
{
    Mask result;
    result.value=_mm512_cmp_ps_mask(value,other.value,_MM_CMPINT_LT);
    return result;
}


//------------------------------------//
//             GREATER THAN              //
//------------------------------------//


template<> inline
typename Number<__m512>::Mask Number<__m512>::operator>(const Number& other) const
{
    Mask result;
    result.value=_mm512_cmp_ps_mask(value,other.value,_MM_CMPINT_GT);
    return result;
}


//------------------------------------//
//             LESS EQUALS              //
//------------------------------------//


template<> inline
typename Number<__m512>::Mask Number<__m512>::operator<=(const Number& other) const
{
    Mask result;
    result.value=_mm512_cmp_ps_mask(value,other.value,_MM_CMPINT_LE);
    return result;
}


//------------------------------------//
//             GREATER EQUALS              //
//------------------------------------//



template<> inline
typename Number<__m512>::Mask Number<__m512>::operator>=(const Number& other) const
{
    Mask result;
    result.value=_mm512_cmp_ps_mask(value,other.value,_MM_CMPINT_GE);
    return result;
}



//------------------------------------//
//               EQUALS               //
//------------------------------------//


template<> inline
typename Number<__m512>::Mask Number<__m512>::operator==(const Number& other) const
{
    Mask result;
    result.value=_mm512_cmp_ps_mask(value,other.value,_MM_CMPINT_EQ);
    return result;
}


//==============================================================//
//                                                              //
//                      BITWISE  OPERATIONS                     //
//                                                              //
//==============================================================//

//------------------------------------//
//             BITWISE AND            //
//------------------------------------//



template<> inline
Number<__m512> Number<__m512>::operator&(const Number& other) const
{Number<__m512> result;result.value=_mm512_castsi512_ps(_mm512_and_epi32(_mm512_castps_si512(value),_mm512_castps_si512(other.value)));return result;}


//------------------------------------//
//             BITWISE OR             //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::operator|(const Number& other) const
{Number<__m512> result;result.value=_mm512_castsi512_ps(_mm512_or_epi32(_mm512_castps_si512(value),_mm512_castps_si512(other.value)));return result;}


//------------------------------------//
//             BITWISE XOR            //
//------------------------------------//



template<> inline
Number<__m512> Number<__m512>::operator^(const Number& other) const
{Number<__m512> result;result.value=_mm512_castsi512_ps(_mm512_xor_epi32(_mm512_castps_si512(value),_mm512_castps_si512(other.value)));return result;}


//------------------------------------//
//             BITWISE NOT            //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::operator~() const
{
    Number<__m512> result;
    floatConverter X;
    X.i = 0xFFFFFFFF;
    BUILD_CONSTANT_16(__one, X.f);
    __m512 val2=_mm512_load_ps(__one);
    result.value=_mm512_castsi512_ps(_mm512_andnot_epi32(_mm512_castps_si512(value),_mm512_castps_si512(val2)));
    return result;
}


//------------------------------------//
//             BITWISE ANDNOT         //
//------------------------------------//



template<> inline
Number<__m512> Number<__m512>::andnot(const Number& other) const
{
    Number<__m512> result;
    __m512i A, B;
    A = _mm512_castps_si512(value);
    B = _mm512_castps_si512(other.value);
    result.value=_mm512_castsi512_ps(_mm512_and_epi32( B, _mm512_xor_epi32( B, A )));
    return result;
}



//==============================================================//
//==============================================================//

//==============================================================//
//                                                              //
//                  ADVANCED OPERATIONS                         //
//                                                              //
//==============================================================//

//------------------------------------//
//              SQUARE ROOT           //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::sqrt() const
{Number<__m512> result;result.value=_mm512_sqrt_ps(value);return result;}


//------------------------------------//
//       RECIPROCAL SQUARE ROOT       //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::rsqrt() const
{
    Number<__m512> result;
    result.value=_mm512_rsqrt23_ps(value);
    return result;
}



//------------------------------------//
//                 LOG                //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::log() const
{
    Number<__m512> result;
    result.value = _mm512_log_ps(value);
return result;
}



//------------------------------------//
//                 EXP                //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::exp() const
{
    Number<__m512> result;
    result.value = _mm512_exp_ps(value);
    return result;
}



//------------------------------------//
//        INVERSE (RECIPROCAL)        //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::inverse() const
{
   Number<__m512> result;
   result.value = _mm512_rcp23_ps( value );
   return result;
}



//------------------------------------//
//        Absolulte Value             //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::abs() const
{
    Number<__m512> neg,result;
    neg.value=_mm512_sub_ps(neg.value,value);
    result.value=_mm512_max_ps(value,neg.value);
    return result;
}



//------------------------------------//
//               SIGN                 //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::sign() const
{
    Number<__m512> zero,poss,result;
    Mask negs;
    BUILD_CONSTANT_16(__one,1.0f);
    __m512 none=_mm512_load_ps(__one);
    none = _mm512_sub_ps(zero.value,none);
    __m512 one=_mm512_load_ps(__one);
    negs.value = _mm512_cmp_ps_mask(value,zero.value,_MM_CMPINT_GE);
    result.value = _mm512_mask_blend_ps(negs.value,none,one);
    
    return result;
}



//------------------------------------//
//               MINIMUM              //
//------------------------------------//


template<> inline
Number<__m512> min(const Number<__m512>& A, const Number<__m512>& B)
{Number<__m512> result;result.value=_mm512_min_ps(A.value, B.value);return result;}


//------------------------------------//
//               MAXIMUM              //
//------------------------------------//


template<> inline
Number<__m512> max(const Number<__m512>& A, const Number<__m512>& B)
{Number<__m512> result;result.value=_mm512_max_ps(A.value, B.value);return result;}



//------------------------------------//
//      Masked assignment (blend)     //
//------------------------------------//


template<> inline
Number<__m512> blend(const typename Number<__m512>::Mask& mask, const Number<__m512>& A, const Number<__m512>& B)
{Number<__m512> result;result.value=_mm512_mask_blend_ps(mask.value,A.value,B.value);return result;}


//------------------------------------//
//      Masked assignment (mask)     //
//------------------------------------//


template<> inline
Number<__m512> Number<__m512>::mask(const Mask& mask) const
{
    Number<__m512> result;
/*
    result.value=_mm512_castsi512_ps(_mm512_mask_xor_epi32(_mm512_castps_si512(result.value),
                                                           mask.value,
                                                           _mm512_castps_si512(value),
                                                           _mm512_castps_si512(value)
                                                           )
                                                           );
*/
    result.value = _mm512_mask_mov_ps( result.value,  mask.value,  value );

    return result;
}



//==============================================================//
//==============================================================//

#ifdef ENABLE_IO_SUPPORT
//==============================================================//
//                                                              //
//                  DEBUG PRINT SUPPORT                         //
//                                                              //
//==============================================================//


template<> inline
std:: ostream & operator<<( std:: ostream & os, const Number<__m512>& number)
{
    float *fp = (float*)&number.value;
    os << "[ " 
       << *(fp+15)
       << " " << *(fp+14)
       << " " << *(fp+13)
       << " " << *(fp+12)
       << " " << *(fp+11)
       << " " << *(fp+10)
       << " " << *(fp+9)
       << " " << *(fp+8)
       << " " << *(fp+7)
       << " " << *(fp+6)
       << " " << *(fp+5)
       << " " << *(fp+4)
       << " " << *(fp+3)
       << " " << *(fp+2)
       << " " << *(fp+1)
       << " " << *(fp) << " ]";
    return os;
}


//==============================================================//
//==============================================================//
#endif


//==============================================================//
//                                                              //
//                  LOADS AND STORES                            //
//                                                              //
//==============================================================//

//------------------------------------//
//              LOADS                 //
//------------------------------------//



template<> inline
void Number<__m512>::Load(const float* data)
{value=_mm512_load_ps((void*)data);}




template<> inline
void Number<__m512>::Load(const float& data)
{value=_mm512_load_ps((void*)&data);}

//------------------------------------//
//              LOADS                 //
//------------------------------------//



template<> inline
void Number<__m512>::Gather(const float* data,const int* offsets)
{
    __m512i index = _mm512_load_epi32((void*)offsets);
    value = _mm512_i32gather_ps(index, (void*)data, 4 );
}




template<> inline
void Number<__m512>::Gather(const float* data,const int& offsets)
{
    __m512i index = _mm512_load_epi32((void*)&offsets);
    value = _mm512_i32gather_ps(index, (void*)data, 4 );
}



//------------------------------------//
//           ALIGNED LOADS            //
//------------------------------------//


template<> inline
void Number<__m512>::Load_Aligned(const float* data)
{value=_mm512_load_ps((void*)data);}





template<> inline
void Number<__m512>::Load_Aligned(const float& data)
{value=_mm512_load_ps((void*)&data);}



//------------------------------------//
//             STORES                 //
//------------------------------------//


template<> inline
void Store(float* data,const Number<__m512>& number)
{_mm512_store_ps(data,number.value);}



template<> inline
void Store(float& data,const Number<__m512>& number)
{_mm512_store_ps(&data,number.value);}



//==============================================================//
//==============================================================//



#endif
