//#####################################################################
//  Copyright (c) 2011-2015 Anonymous Authors of Paper Submission paper0191.
//  This file is covered by the FreeBSD license. Please refer to the 
//  license.txt file for more information.
//#####################################################################


#ifndef SUBROUTINE_Force_Stabilization
#include <assert.h>
#include "KernelCommon.h"
#else
namespace {
#endif

    namespace nm_Force_Stabilization{
        BUILD_CONSTANT(minusseven,-7.f);
        BUILD_CONSTANT(three,3.f);
        BUILD_CONSTANT(minusfive,-5.f);
        BUILD_CONSTANT(one,1.f);
    }
    
    using namespace nm_Force_Stabilization;

    
    template<class Tw,class T_DATA=void,class I_DATA=void>
#ifdef SUBROUTINE_Force_Stabilization
    inline
#endif
    void Force_Stabilization(const T_DATA (&Du)[3][8], const T_DATA (&constant), T_DATA (&dH)[3][8] )
    {
        typedef Number<Tw> Tn;

        const int OpOrder[8][8] = {
            {0, 1, 2, 4, 3, 5, 6, 7},
            {1, 0, 3, 5, 2, 4, 7, 6},
            {2, 3, 0, 6, 1, 7, 4, 5},
            {3, 2, 1, 7, 0, 6, 5, 4},
            {4, 5, 6, 0, 7, 1, 2, 3},
            {5, 4, 7, 1, 6, 0, 3, 2},
            {6, 7, 4, 2, 5, 3, 0, 1},
            {7, 6, 5, 3, 4, 2, 1, 0}};

        Tn Ru[8];
        Tn Rf[8];
      
        Tn TEMP;
        Tn Rminusseven;
        Tn Rthree;
        Tn Rone;
        Tn Rminusfive;
        Tn RConstant;
        Tn RWeight;
        
        Rminusseven.Load_Aligned(minusseven);
        Rthree.Load_Aligned(three);
        Rone.Load_Aligned(one);
        Rminusfive.Load_Aligned(minusfive);
        
        RConstant.Load( constant );
        /*RWeight.Load( weight );*/
        RConstant = RConstant;// * RWeight;

        for( int d = 0; d < 3; d++ )
            {
                for( int i = 0; i < 8; i++)
                    Ru[i].Load(Du[d][i]);
                
                for( int i = 0; i < 8; i++)
                    {
                        Rf[i] = 
                            (Rminusseven * Ru[OpOrder[i][0]]) + 
                            (Rthree * Ru[OpOrder[i][1]])+
                            (Rthree * Ru[OpOrder[i][2]])+
                            (Rthree * Ru[OpOrder[i][3]])+
                            (Rone * Ru[OpOrder[i][4]])+
                            (Rone * Ru[OpOrder[i][5]])+
                            (Rone * Ru[OpOrder[i][6]])+
                            (Rminusfive * Ru[OpOrder[i][7]]);
                    }

                for( int i = 0; i < 8; i++)
                    {
                        Ru[i].Load(dH[d][i]);
                        Rf[i] = Rf[i] * RConstant;
                        Rf[i] = Rf[i] + Ru[i];
                        Store( dH[d][i], Rf[i]);
                    }
            }

    }

#ifdef SUBROUTINE_Force_Stabilization
}
#else
#define INSTANCE_KERNEL_Force_Stabilization(WIDTH) const WIDETYPE(float,WIDTH) (&Du)[3][8], const WIDETYPE(float,WIDTH) (&constant),  WIDETYPE(float,WIDTH) (&dH)[3][8]
INSTANCE_KERNEL(Force_Stabilization);
#undef INSTANCE_KERNEL_Force_Stabilization
#endif
