
//#######################################################################
//##  Copyright (c) 2011-2015 Anonymous Authors of Paper Submission paper0191.
//##  This file is covered by the FreeBSD license. Please refer to the 
//##  license.txt file for more information.
//#######################################################################


#include <cmath>
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <sys/time.h>
#include "KernelCommon.h"
struct NEOHOOKEAN_TAG;
struct COROTATED_TAG;

#include "Force_Stabilization.h"

#include <Thread_Queueing/PTHREAD_QUEUE.h>
#include <Kernel_Serial_Base_Helper.h>


template < class T > T Get_Random (const T a = (T) - 1., const T b = (T) 1.)
{
  return ((b - a) * (T) rand ()) / (T) RAND_MAX + a;
}

struct timeval starttime, stoptime;
void
start_timer ()
{
  gettimeofday (&starttime, NULL);
}

void
stop_timer ()
{
  gettimeofday (&stoptime, NULL);
}

double
get_time ()
{
  return (double) stoptime.tv_sec - (double) starttime.tv_sec +
    (double) 1e-6 *(double) stoptime.tv_usec -
    (double) 1e-6 *(double) starttime.tv_usec;
}


template < int SIZE > class Force_Stabilization_SCALAR_None
{
private:
  // Generate Variables Here
  float *_local_Du;
  float *_local_constant;
  float *_local_dH;

public:
    explicit Force_Stabilization_SCALAR_None (float *Du_in, float *constant_in,
                                              float *dH_in):_local_Du (Du_in),
    _local_constant (constant_in), _local_dH (dH_in)
  {
  }
  void Execute (int index)
  {
    // full array typedefs
    //typedef int (&refArray)[SIZE][3][8];
    typedef float (&fullArray1)[SIZE][3][8][16];
    typedef float (&fullArray2)[SIZE][16];
    typedef float (&fullArray3)[SIZE][3][8][16];

    // full array extractions
    //refArray _rA = reinterpret_cast < refArray >(*_A);
    fullArray1 _rDu = reinterpret_cast < fullArray1 > (*_local_Du);
    fullArray2 _rconstant = reinterpret_cast < fullArray2 > (*_local_constant);
    fullArray3 _rdH = reinterpret_cast < fullArray3 > (*_local_dH);

    const int ChunkSize = 1;
    // chunk typedef
    //typedef int (&refArray1)[3][8];
    typedef float (&refArray1)[3][8][16];
    typedef float (&refArray2)[16];
    typedef float (&refArray3)[3][8][16];

    for (int chunk_offset = 0; chunk_offset < 16; chunk_offset += ChunkSize)
      {
        //refArray1 Ak = reinterpret_cast < refArray1 > (_rA[entry][0][chunk_offset]);
        refArray1 Duk =
          reinterpret_cast < refArray1 > (_rDu[index][0][0][chunk_offset]);
        refArray2 constantk =
          reinterpret_cast < refArray2 > (_rconstant[index][chunk_offset]);
        refArray3 dHk =
          reinterpret_cast < refArray3 > (_rdH[index][0][0][chunk_offset]);

        Force_Stabilization < float, float[16], int[16] > (Duk, constantk, dHk);
      }

  }
};


#ifdef ENABLE_SSE_INSTRUCTION_SET

template < int SIZE > class Force_Stabilization_SSE_None
{
private:
  // Generate Variables Here
  float *_local_Du;
  float *_local_constant;
  float *_local_dH;

public:
    explicit Force_Stabilization_SSE_None (float *Du_in, float *constant_in,
                                           float *dH_in):_local_Du (Du_in),
    _local_constant (constant_in), _local_dH (dH_in)
  {
  }
  void Execute (int index)
  {
    // full array typedefs
    //typedef int (&refArray)[SIZE][3][8];
    typedef float (&fullArray1)[SIZE][3][8][16];
    typedef float (&fullArray2)[SIZE][16];
    typedef float (&fullArray3)[SIZE][3][8][16];

    // full array extractions
    //refArray _rA = reinterpret_cast < refArray >(*_A);
    fullArray1 _rDu = reinterpret_cast < fullArray1 > (*_local_Du);
    fullArray2 _rconstant = reinterpret_cast < fullArray2 > (*_local_constant);
    fullArray3 _rdH = reinterpret_cast < fullArray3 > (*_local_dH);

    const int ChunkSize = 4;
    // chunk typedef
    //typedef int (&refArray1)[3][8];
    typedef float (&refArray1)[3][8][16];
    typedef float (&refArray2)[16];
    typedef float (&refArray3)[3][8][16];

    for (int chunk_offset = 0; chunk_offset < 16; chunk_offset += ChunkSize)
      {
        //refArray1 Ak = reinterpret_cast < refArray1 > (_rA[entry][0][chunk_offset]);
        refArray1 Duk =
          reinterpret_cast < refArray1 > (_rDu[index][0][0][chunk_offset]);
        refArray2 constantk =
          reinterpret_cast < refArray2 > (_rconstant[index][chunk_offset]);
        refArray3 dHk =
          reinterpret_cast < refArray3 > (_rdH[index][0][0][chunk_offset]);

        Force_Stabilization < __m128, float[16], int[16] > (Duk, constantk,
                                                            dHk);
      }

  }
};

#endif

#ifdef ENABLE_AVX_INSTRUCTION_SET

template < int SIZE > class Force_Stabilization_AVX_None
{
private:
  // Generate Variables Here
  float *_local_Du;
  float *_local_constant;
  float *_local_dH;

public:
    explicit Force_Stabilization_AVX_None (float *Du_in, float *constant_in,
                                           float *dH_in):_local_Du (Du_in),
    _local_constant (constant_in), _local_dH (dH_in)
  {
  }
  void Execute (int index)
  {
    // full array typedefs
    //typedef int (&refArray)[SIZE][3][8];
    typedef float (&fullArray1)[SIZE][3][8][16];
    typedef float (&fullArray2)[SIZE][16];
    typedef float (&fullArray3)[SIZE][3][8][16];

    // full array extractions
    //refArray _rA = reinterpret_cast < refArray >(*_A);
    fullArray1 _rDu = reinterpret_cast < fullArray1 > (*_local_Du);
    fullArray2 _rconstant = reinterpret_cast < fullArray2 > (*_local_constant);
    fullArray3 _rdH = reinterpret_cast < fullArray3 > (*_local_dH);

    const int ChunkSize = 8;
    // chunk typedef
    //typedef int (&refArray1)[3][8];
    typedef float (&refArray1)[3][8][16];
    typedef float (&refArray2)[16];
    typedef float (&refArray3)[3][8][16];

    for (int chunk_offset = 0; chunk_offset < 16; chunk_offset += ChunkSize)
      {
        //refArray1 Ak = reinterpret_cast < refArray1 > (_rA[entry][0][chunk_offset]);
        refArray1 Duk =
          reinterpret_cast < refArray1 > (_rDu[index][0][0][chunk_offset]);
        refArray2 constantk =
          reinterpret_cast < refArray2 > (_rconstant[index][chunk_offset]);
        refArray3 dHk =
          reinterpret_cast < refArray3 > (_rdH[index][0][0][chunk_offset]);

        Force_Stabilization < __m256, float[16], int[16] > (Duk, constantk,
                                                            dHk);
      }

  }
};

#endif

#ifdef ENABLE_NEON_INSTRUCTION_SET

template < int SIZE > class Force_Stabilization_NEON_None
{
private:
  // Generate Variables Here
  float *_local_Du;
  float *_local_constant;
  float *_local_dH;

public:
    explicit Force_Stabilization_NEON_None (float *Du_in, float *constant_in,
                                            float *dH_in):_local_Du (Du_in),
    _local_constant (constant_in), _local_dH (dH_in)
  {
  }
  void Execute (int index)
  {
    // full array typedefs
    //typedef int (&refArray)[SIZE][3][8];
    typedef float (&fullArray1)[SIZE][3][8][16];
    typedef float (&fullArray2)[SIZE][16];
    typedef float (&fullArray3)[SIZE][3][8][16];

    // full array extractions
    //refArray _rA = reinterpret_cast < refArray >(*_A);
    fullArray1 _rDu = reinterpret_cast < fullArray1 > (*_local_Du);
    fullArray2 _rconstant = reinterpret_cast < fullArray2 > (*_local_constant);
    fullArray3 _rdH = reinterpret_cast < fullArray3 > (*_local_dH);

    const int ChunkSize = 4;
    // chunk typedef
    //typedef int (&refArray1)[3][8];
    typedef float (&refArray1)[3][8][16];
    typedef float (&refArray2)[16];
    typedef float (&refArray3)[3][8][16];

    for (int chunk_offset = 0; chunk_offset < 16; chunk_offset += ChunkSize)
      {
        //refArray1 Ak = reinterpret_cast < refArray1 > (_rA[entry][0][chunk_offset]);
        refArray1 Duk =
          reinterpret_cast < refArray1 > (_rDu[index][0][0][chunk_offset]);
        refArray2 constantk =
          reinterpret_cast < refArray2 > (_rconstant[index][chunk_offset]);
        refArray3 dHk =
          reinterpret_cast < refArray3 > (_rdH[index][0][0][chunk_offset]);

        Force_Stabilization < float32x4_t, float[16], int[16] > (Duk, constantk,
                                                                 dHk);
      }

  }
};

#endif

#ifdef ENABLE_MIC_INSTRUCTION_SET

template < int SIZE > class Force_Stabilization_MIC_None
{
private:
  // Generate Variables Here
  float *_local_Du;
  float *_local_constant;
  float *_local_dH;

public:
    explicit Force_Stabilization_MIC_None (float *Du_in, float *constant_in,
                                           float *dH_in):_local_Du (Du_in),
    _local_constant (constant_in), _local_dH (dH_in)
  {
  }
  void Execute (int index)
  {
    // full array typedefs
    //typedef int (&refArray)[SIZE][3][8];
    typedef float (&fullArray1)[SIZE][3][8][16];
    typedef float (&fullArray2)[SIZE][16];
    typedef float (&fullArray3)[SIZE][3][8][16];

    // full array extractions
    //refArray _rA = reinterpret_cast < refArray >(*_A);
    fullArray1 _rDu = reinterpret_cast < fullArray1 > (*_local_Du);
    fullArray2 _rconstant = reinterpret_cast < fullArray2 > (*_local_constant);
    fullArray3 _rdH = reinterpret_cast < fullArray3 > (*_local_dH);

    const int ChunkSize = 16;
    // chunk typedef
    //typedef int (&refArray1)[3][8];
    typedef float (&refArray1)[3][8][16];
    typedef float (&refArray2)[16];
    typedef float (&refArray3)[3][8][16];

    for (int chunk_offset = 0; chunk_offset < 16; chunk_offset += ChunkSize)
      {
        //refArray1 Ak = reinterpret_cast < refArray1 > (_rA[entry][0][chunk_offset]);
        refArray1 Duk =
          reinterpret_cast < refArray1 > (_rDu[index][0][0][chunk_offset]);
        refArray2 constantk =
          reinterpret_cast < refArray2 > (_rconstant[index][chunk_offset]);
        refArray3 dHk =
          reinterpret_cast < refArray3 > (_rdH[index][0][0][chunk_offset]);

        Force_Stabilization < __m512, float[16], int[16] > (Duk, constantk,
                                                            dHk);
      }

  }
};

#endif

int
main (int argc, char *argv[])
{
  typedef float T;

  int seed = 1;
  int threads = 1;
  int threads_max = 1;
  int passes = 1;
  const int data_size = 1000000;
  if (argc >= 2)
    threads = atoi (argv[1]);
  if (argc >= 3)
    threads_max = atoi (argv[2]);
  if (argc >= 4)
    passes = atoi (argv[3]);
  srand (seed);

  pthread_queue = new PTHREAD_QUEUE (threads_max);

  std::
    cout << "Preparing to Run " << data_size << " of all kernels with " <<
    threads << " threads." << std::endl;



  {
    std::cout << "Running Thread Test for Force_Stabilization " << std::endl;

//=======================================================
//
//        DEFINE ALL VARIABLES USED BY KERNEL
//
//=======================================================
    std::cout << "\nAllocating all data: ";
    std::cout.flush ();

    start_timer ();
    typedef T (&Du_type)[data_size][3][8][16];
    Du_type Du =
      reinterpret_cast < Du_type >
      (*((T *) (_mm_malloc (data_size * 3 * 8 * 16 * sizeof (T), 64))));
    typedef T (&constant_type)[data_size][16];
    constant_type constant =
      reinterpret_cast < constant_type >
      (*((T *) (_mm_malloc (data_size * 16 * sizeof (T), 64))));
    typedef T (&dH_type)[data_size][3][8][16];
    dH_type dH =
      reinterpret_cast < dH_type >
      (*((T *) (_mm_malloc (data_size * 3 * 8 * 16 * sizeof (T), 64))));


    for (int __a = 0; __a < data_size; __a++)
      for (int __b = 0; __b < 3; __b++)
        for (int __c = 0; __c < 8; __c++)
          for (int __d = 0; __d < 16; __d++)
            {
              Du[__a][__b][__c][__d] = Get_Random < float >();
    } for (int __a = 0; __a < data_size; __a++)
      for (int __b = 0; __b < 16; __b++)
        {
          constant[__a][__b] = Get_Random < float >();
    } for (int __a = 0; __a < data_size; __a++)
      for (int __b = 0; __b < 3; __b++)
        for (int __c = 0; __c < 8; __c++)
          for (int __d = 0; __d < 16; __d++)
            {
              dH[__a][__b][__c][__d] = Get_Random < float >();
            }
    stop_timer ();

    std::cout << get_time () << "s\n\n" << std::endl;


//=======================================================
//
//             COMPUTE SCALAR RESULTS
//
//=======================================================

    {
      std::cout << "	Running " << data_size << " of SCALAR :  " << std::endl;


      Force_Stabilization_SCALAR_None < data_size > op ((float *) &Du,
                                                        (float *) &constant,
                                                        (float *) &dH);

      for (int t = threads; t < threads_max;
           t += std::max < int >(((threads_max - threads) / 30), 1))
        {
          std::cout << "Running Test with " << t << " threads." << std::endl;
          MT_Streaming_Kernels::Kernel_Serial_Base_Helper <
            Force_Stabilization_SCALAR_None < data_size > >helper (op,
                                                                   data_size,
                                                                   t);

          double min_time = 10000000;
          double max_time = -1;
          double avg_time = 0;

          for (int i = 0; i < passes; i++)
            {
              start_timer ();
              helper.Run_Parallel ();
              stop_timer ();
              std::cout << get_time () << "s" << std::endl;
              min_time = std::min < double >(min_time, get_time ());
              max_time = std::max < double >(max_time, get_time ());
              avg_time += get_time ();
            }
          avg_time = avg_time / passes;
          std::cout << "Min pass time: " << min_time << std::endl;
          std::cout << "Max pass time: " << max_time << std::endl;
          std::cout << "Avg pass time: " << avg_time << std::endl;
        }




    }

//=======================================================
//
//             COMPUTE SSE RESULTS
//
//=======================================================
#ifdef ENABLE_SSE_INSTRUCTION_SET
    {
      std::cout << "	Running " << data_size << " of SSE :  " << std::endl;


      Force_Stabilization_SSE_None < data_size > op ((float *) &Du,
                                                     (float *) &constant,
                                                     (float *) &dH);

      for (int t = threads; t < threads_max;
           t += std::max < int >(((threads_max - threads) / 30), 1))
        {
          std::cout << "Running Test with " << t << " threads." << std::endl;
          MT_Streaming_Kernels::Kernel_Serial_Base_Helper <
            Force_Stabilization_SSE_None < data_size > >helper (op, data_size,
                                                                t);

          double min_time = 10000000;
          double max_time = -1;
          double avg_time = 0;

          for (int i = 0; i < passes; i++)
            {
              start_timer ();
              helper.Run_Parallel ();
              stop_timer ();
              std::cout << get_time () << "s" << std::endl;
              min_time = std::min < double >(min_time, get_time ());
              max_time = std::max < double >(max_time, get_time ());
              avg_time += get_time ();
            }
          avg_time = avg_time / passes;
          std::cout << "Min pass time: " << min_time << std::endl;
          std::cout << "Max pass time: " << max_time << std::endl;
          std::cout << "Avg pass time: " << avg_time << std::endl;
        }




    }
#endif

//=======================================================
//
//             COMPUTE AVX RESULTS
//
//=======================================================
#ifdef ENABLE_AVX_INSTRUCTION_SET
    {
      std::cout << "	Running " << data_size << " of AVX :  " << std::endl;


      Force_Stabilization_AVX_None < data_size > op ((float *) &Du,
                                                     (float *) &constant,
                                                     (float *) &dH);

      for (int t = threads; t < threads_max;
           t += std::max < int >(((threads_max - threads) / 30), 1))
        {
          std::cout << "Running Test with " << t << " threads." << std::endl;
          MT_Streaming_Kernels::Kernel_Serial_Base_Helper <
            Force_Stabilization_AVX_None < data_size > >helper (op, data_size,
                                                                t);

          double min_time = 10000000;
          double max_time = -1;
          double avg_time = 0;

          for (int i = 0; i < passes; i++)
            {
              start_timer ();
              helper.Run_Parallel ();
              stop_timer ();
              std::cout << get_time () << "s" << std::endl;
              min_time = std::min < double >(min_time, get_time ());
              max_time = std::max < double >(max_time, get_time ());
              avg_time += get_time ();
            }
          avg_time = avg_time / passes;
          std::cout << "Min pass time: " << min_time << std::endl;
          std::cout << "Max pass time: " << max_time << std::endl;
          std::cout << "Avg pass time: " << avg_time << std::endl;
        }




    }
#endif

//=======================================================
//
//             COMPUTE NEON RESULTS
//
//=======================================================
#ifdef ENABLE_NEON_INSTRUCTION_SET
    {
      std::cout << "	Running " << data_size << " of NEON :  " << std::endl;


      Force_Stabilization_NEON_None < data_size > op ((float *) &Du,
                                                      (float *) &constant,
                                                      (float *) &dH);

      for (int t = threads; t < threads_max;
           t += std::max < int >(((threads_max - threads) / 30), 1))
        {
          std::cout << "Running Test with " << t << " threads." << std::endl;
          MT_Streaming_Kernels::Kernel_Serial_Base_Helper <
            Force_Stabilization_NEON_None < data_size > >helper (op, data_size,
                                                                 t);

          double min_time = 10000000;
          double max_time = -1;
          double avg_time = 0;

          for (int i = 0; i < passes; i++)
            {
              start_timer ();
              helper.Run_Parallel ();
              stop_timer ();
              std::cout << get_time () << "s" << std::endl;
              min_time = std::min < double >(min_time, get_time ());
              max_time = std::max < double >(max_time, get_time ());
              avg_time += get_time ();
            }
          avg_time = avg_time / passes;
          std::cout << "Min pass time: " << min_time << std::endl;
          std::cout << "Max pass time: " << max_time << std::endl;
          std::cout << "Avg pass time: " << avg_time << std::endl;
        }




    }
#endif

//=======================================================
//
//             COMPUTE MIC RESULTS
//
//=======================================================
#ifdef ENABLE_MIC_INSTRUCTION_SET
    {
      std::cout << "	Running " << data_size << " of MIC :  " << std::endl;


      Force_Stabilization_MIC_None < data_size > op ((float *) &Du,
                                                     (float *) &constant,
                                                     (float *) &dH);

      for (int t = threads; t < threads_max;
           t += std::max < int >(((threads_max - threads) / 30), 1))
        {
          std::cout << "Running Test with " << t << " threads." << std::endl;
          MT_Streaming_Kernels::Kernel_Serial_Base_Helper <
            Force_Stabilization_MIC_None < data_size > >helper (op, data_size,
                                                                t);

          double min_time = 10000000;
          double max_time = -1;
          double avg_time = 0;

          for (int i = 0; i < passes; i++)
            {
              start_timer ();
              helper.Run_Parallel ();
              stop_timer ();
              std::cout << get_time () << "s" << std::endl;
              min_time = std::min < double >(min_time, get_time ());
              max_time = std::max < double >(max_time, get_time ());
              avg_time += get_time ();
            }
          avg_time = avg_time / passes;
          std::cout << "Min pass time: " << min_time << std::endl;
          std::cout << "Max pass time: " << max_time << std::endl;
          std::cout << "Avg pass time: " << avg_time << std::endl;
        }




    }
#endif

//=======================================================
//
//        FREE MEMORY USED BY ALL VARIABLES
//
//=======================================================
    std::cout << "\nFreeing all data: " << std::endl;
    std::cout.flush ();

    _mm_free (reinterpret_cast < void *>(Du));
    _mm_free (reinterpret_cast < void *>(constant));
    _mm_free (reinterpret_cast < void *>(dH));


  }


  return 0;

}
